from importlib import import_module
import logging
import sys
import os
from arkitekt.apps.fakts import ArkitektFakts
from fakts import Fakts
from fakts.grants.remote.static import StaticGrant
from fakts.discovery.static import StaticDiscovery
from rich.console import Console
from arkitekt import Arkitekt

logger = logging.getLogger(__name__)


class Run:
    def __init__(
        self,
        path=None,
        entrypoint="run",
    ) -> None:

        self.console = Console()
        if path == ".":
            self.watch_path = os.getcwd()
            sys.path.insert(0, self.watch_path)
        else:
            self.watch_path = os.path.join(os.getcwd(), path)

        self.console.print(f"watch_path: {self.watch_path}")

        self.module_path = f"{path}.{entrypoint}" if path != "." else f"{entrypoint}"
        self.module = import_module(self.module_path)
        os.environ["ARKITEKT_AGENT_DEBUG"] = "False"

    async def run(self, token=None, endpoint=None):

        app = Arkitekt(
            fakts=ArkitektFakts(
                grant=StaticGrant(
                    discovery=StaticDiscovery(
                        base_url=endpoint or os.getenv("FAKTS_ENDPOINT_URL")
                    ),
                    token=token or os.getenv("FAKTS_TOKEN"),
                )
            )
        )

        async with app:
            await app.rekuest.run()


async def import_directory_and_start(path, token=None, endpoint=None):
    host = Run(path=path)
    await host.run(token=token, endpoint=endpoint)



import asyncio

import aiohttp
from rich.console import Console
from arkitekt.apps.connected import App


async def check_fakts(endpoint_url: str):

    async with aiohttp.ClientSession(
        headers={"Content-Type": "application/json"}
    ) as session:
        # get json from endpoint
        async with session.get(endpoint_url) as resp:

            healthz_json = await resp.text()
            return healthz_json


async def check_app(app: App):

    async with app:
        check_futures = []
        check_keys = []

        for key, service in app:
            if hasattr(service, "healthz"):
                check_futures.append(service.healthz.check())
                check_keys.append(key)

        return dict(zip(check_keys, await asyncio.gather(*check_futures)))


async def check_app_loop(
    console: Console, app: App, interval: int = 5, retries: int = 0
):

    while True:
        try:
            x = await check_app(app)
            return x

        except:
            if retries == 0:
                console.print(f"Failed to connect to {app}")
                raise

            else:
                console.print_exception()
                console.print("Retrying...")
                await asyncio.sleep(interval)
                retries -= 1


async def check_fakts_loop(
    console: Console, endpoint_url: str, interval: int = 5, retries: int = 0
):

    while True:
        try:
            x = await check_fakts(endpoint_url)
            return x

        except:
            if retries == 0:
                console.print(f"Failed to connect to {endpoint_url}")
                raise

            else:
                console.print_exception()
                console.print("Retrying...")
                await asyncio.sleep(interval)
                retries -= 1



import argparse
import asyncio
import os
import sys
from enum import Enum

from rich.console import Console
from rich.prompt import Confirm, Prompt
from rich.table import Table

from arkitekt.apps.connected import App
from arkitekt.cli.prod.check import check_app, check_app_loop, check_fakts_loop
from arkitekt.cli.prod.dump import import_directory_and_dump
from arkitekt.cli.prod.run import import_directory_and_start
from fakts import Fakts
from fakts.discovery.advertised import AdvertisedDiscovery
from fakts.discovery.static import StaticDiscovery
from fakts.grants.remote.device_code import DeviceCodeGrant
from fakts.grants.remote.static import StaticGrant
from herre import Herre
from arkitekt.cli.logic.initialize import initialize_manifest
from arkitekt.cli.logic.connect import connect
from arkitekt.cli.logic.run import run
from arkitekt.cli.logic.dev import dev
from arkitekt.cli.logic.build import build
import yaml


try:
    from rich.traceback import install

    install()
except:
    pass

directory = os.getcwd()


class ArkitektOptions(str, Enum):
    INIT = "init"
    CONNECT = "connect"
    RUN = "run"
    WAIT = "wait"
    DEV = "dev"
    DUMP = "dump"
    BUILD = "build"


class TemplateOptions(str, Enum):
    AGENT = "agent"
    CLIENT = "client"


class GrantOptions(str, Enum):
    CLAIM = "claim"
    DEVICE = "device"


agent_script = f'''
from arkitekt import register

@register()
def test_show(name: str)-> str:
    """Demo Node

    This demo node will just pass through
    a name 

    Args:
        name (str): The name

    Returns:
        str: The pass throughed Name
    """
    return name


'''

mikro_script = f"""
from arkitekt import Arkitekt

app = Arkitekt()

with app:
    # Implement your logig here
"""

HEADER = """
                __    _  __         __    __ 
  ____ _ _____ / /__ (_)/ /_ ___   / /__ / /_
 / __ `// ___// //_// // __// _ \ / //_// __/
/ /_/ // /   / ,<  / // /_ /  __// ,<  / /_  
\__,_//_/   /_/|_|/_/ \__/ \___//_/|_| \__/  
                                             
"""


template_map = {"agent": agent_script, "client": mikro_script}


def main(
    script=ArkitektOptions.INIT,
    name=None,
    path=".",
    refresh=False,
    scan_network=None,
    beacon=None,
    template=None,
    services=[],
    claim=False,
    silent=False,
    overwrite=False,
    token=None,
    endpoint_url=None,
    retries=3,
    interval=1,
    tag=None,
    grant=GrantOptions.DEVICE,
):

    console = Console()
    console.print(HEADER)

    try:

        if path == ".":
            app_directory = os.getcwd()
            name = name or os.path.basename(app_directory)
        else:
            app_directory = os.path.join(os.getcwd(), path)
            name = name or path

        run_script_path = os.path.join(app_directory, "run.py")
        config_path = os.path.join(app_directory, ".arkitekt")

        if script == ArkitektOptions.INIT:
            initialize_manifest(app_directory=app_directory, silent=silent)
            return

        if script == ArkitektOptions.CONNECT:
            connect(app_directory=app_directory, silent=silent)
            return

        if script == ArkitektOptions.RUN:
            run(path=path, silent=silent, token=token, endpoint=endpoint_url)
            return

        if script == ArkitektOptions.DEV:
            dev(path=path, silent=silent, token=token, endpoint=endpoint_url)
            return

        if script == ArkitektOptions.BUILD:
            build(path=path, silent=silent, tag=tag)
            return

        if script == ArkitektOptions.WAIT:

            url = endpoint_url or (
                Prompt.ask(
                    "Give us the path to your fakts instance",
                    default=os.getenv("FAKTS_ENDPOINT_URL", "http://localhost:8000/f/"),
                )
                if not silent
                else os.getenv("FAKTS_ENDPOINT_URL", "http://localhost:8000/f/")
            )
            answer = asyncio.run(
                check_fakts_loop(
                    console,
                    endpoint_url=url,
                    retries=retries,
                    interval=interval,
                )
            )
            if answer:
                console.print("Fakts is up and running")
            return

        if script == ArkitektOptions.DEV:

            from arkitekt.cli.dev.autostage import watch_directory_and_stage

            if not os.path.isfile(run_script_path):
                console.print(
                    f"Could't find a run.py in {app_directory} {run_script_path}"
                )
                return
            if not os.path.isfile(fakts_path):
                console.print(f"{app_directory} does not have a valid fakts.yaml")
                return

            asyncio.run(watch_directory_and_stage(path, entrypoint="run"))

        if script == ArkitektOptions.RUN:

            if not os.path.isfile(run_script_path):
                console.print(
                    f"Could't find a run.py in {app_directory} {run_script_path}"
                )
                return
            if not os.path.isfile(fakts_path):
                console.print(f"{app_directory} does not have a valid fakts.yaml")
                return

            asyncio.run(import_directory_and_start(path, entrypoint="run"))

        if script == ArkitektOptions.DUMP:

            if not os.path.isfile(run_script_path):
                console.print(
                    f"Could't find a run.py in {app_directory} {run_script_path}"
                )
                return
            if not os.path.isfile(fakts_path):
                console.print(f"{app_directory} does not have a valid fakts.yaml")
                return

            asyncio.run(import_directory_and_dump(path, entrypoint="run"))

        if script == ArkitektOptions.CHECK:

            if not os.path.isfile(fakts_path):
                console.print(
                    f"Directory does not containt a valid fakts.yaml. Please initialize through 'arkitekt init {path}' first"
                )
                return

            fakts = Fakts(fakts_path=fakts_path)
            if not fakts.loaded_fakts:
                console.print(
                    f"Configuration in fakts.yaml is not sufficient please reinitilize through 'arkitekt init {path}'"
                )

            app = App(fakts=fakts)
            state = asyncio.run(
                check_app_loop(
                    console,
                    app,
                    retries=retries,
                    interval=interval,
                )
            )

            table = Table(title="App State")

            table.add_column("Service")
            table.add_column("Subservice")
            table.add_column("Message")

            for service, subservices in state.items():
                for subservice, status in subservices.items():
                    table.add_row(service, subservice, status)

            console.print(table)

        if script == ArkitektOptions.INIT:

            if os.path.exists(app_directory):
                console.print(f"App director already existed")
            else:
                os.mkdir(app_directory)
                console.print(f"Created a new App Directory at {app_directory}")

            initialize_fakts = True
            if os.path.isfile(fakts_path):
                console.print(
                    f"There seems to have been a previous app initialized at {app_directory}"
                )
                if refresh is None:
                    initialize_fakts = (
                        Confirm.ask("Do you want to refresh the Apps configuration? ")
                        if not silent
                        else False
                    )
                    if not initialize_fakts:
                        console.print(f"Nothing done. Bye :smiley: ")
                        return
                else:
                    initialize_fakts = refresh

            if initialize_fakts:

                console.print(f"Initializing a new Configuration for {name}")

                app_identifier = (
                    name if silent else Prompt.ask("Give your app a unique identifier")
                )
                app_version = (
                    "0.0.1" if silent else Prompt.ask("Give your app a version")
                )
                app_description = (
                    "A new app" if silent else Prompt.ask("Give your app a description")
                )

                console.log("Which scopes do you want your app to have?")
                scopes = []
                for scope in available_scopes:
                    if Confirm.ask(f"App should have {scope} scope?"):
                        scopes.append(scope)

                if not scopes:
                    console.print("App needs at least one scope")
                    return

                manifest = {
                    "identifier": app_identifier,
                    "version": app_version,
                    "description": app_description,
                    "scopes": scopes,
                }

                with open(fakts_path, "w") as f:
                    yaml.dump(manifest, f)

                if scan_network is None:
                    scan_network = (
                        Confirm.ask(
                            "Do you want to automatically scan the network for local instances?"
                        )
                        if not silent
                        else False
                    )

                    if scan_network:
                        discovery = AdvertisedDiscovery()
                    else:
                        url = endpoint_url or (
                            Prompt.ask(
                                "Give us the path to your fakts instance",
                                default=os.getenv(
                                    "FAKTS_ENDPOINT_URL", "http://localhost:8000/f/"
                                ),
                            )
                            if not silent
                            else os.getenv(
                                "FAKTS_ENDPOINT_URL", "http://localhost:8000/f/"
                            )
                        )
                        discovery = StaticDiscovery(base_url=url)

                    if grant == GrantOptions.CLAIM:
                        console.print(f"Initializing through claiming process")
                        client_id = client_id or os.getenv("FAKTS_CLIENT_ID", None)
                        client_secret = client_secret or os.getenv(
                            "FAKTS_CLIENT_SECRET", None
                        )

                        assert (
                            client_id and client_secret
                        ), "Please provide a client id and secret or set FAKTS_CLIENT_ID and FAKTS_CLIENT_SECRET as env variables"

                        fakts = Fakts(
                            grant=ClaimGrant(
                                discovery=discovery,
                                client_id=client_id,
                                client_secret=client_secret,
                            ),
                        )

                    elif grant == GrantOptions.DEVICE:
                        console.print(f"Initializing through device code")
                        has_webrowser = (
                            Confirm.ask(
                                "Do you have a webbrowser installed? Otherwise we will generate a code that you need to enter manually"
                            )
                            if not silent
                            else False
                        )

                        fakts = Fakts(
                            grant=DeviceCodeGrant(
                                discovery=discovery, open_browser=has_webrowser
                            ),
                            fakts_path=fakts_path,
                            force_refresh=True,
                        )

                    else:
                        raise NotImplementedError("Grant type not implemented")

                    with fakts as f:
                        f.load()

            save_run = True
            if os.path.isfile(run_script_path):
                save_run = refresh or (
                    Confirm.ask(
                        "run.py already exists? Do you want to overwrite the run.py file? :smiley: "
                    )
                    if not silent
                    else False
                )

            if save_run:
                console.print("Initializing Templates for the App")
                if not template:
                    template = (
                        Prompt.ask(
                            "Which Template do you want to use?",
                            choices=list(template_map.keys()),
                            default="agent",
                        )
                        if not silent
                        else "agent"
                    )

                with open(run_script_path, "w") as f:
                    f.write(template_map[template])
                    console.print("Create new Entrypoint for this App")

    except:
        console.print_exception()
        sys.exit(1)


def entrypoint():
    parser = argparse.ArgumentParser(description="Say hello")
    parser.add_argument("script", type=ArkitektOptions, help="The Script Type")
    parser.add_argument("path", type=str, help="The Path", nargs="?", default=".")
    parser.add_argument("--name", type=str, help="The Name of this script")
    parser.add_argument(
        "--retries", type=int, help="The Name of this script", default=0
    )
    parser.add_argument(
        "-t", type=str, help="The tag of the dockercontainer", default=None
    )
    parser.add_argument(
        "--interval", type=int, help="The Name of this script", default=5
    )
    parser.add_argument("--scan-network", type=bool, help="Do you want to refresh")
    parser.add_argument("--beacon", type=str, help="The adress of the beacon")
    parser.add_argument(
        "--template", type=TemplateOptions, help="The run script template"
    )
    parser.add_argument(
        "--token",
        type=str,
        help="The token to negoatite the connection",
    )
    parser.add_argument(
        "--endpoint_url",
        type=str,
        help="The endpoint url for the fakts instance",
    )
    parser.add_argument(
        "--grant", type=GrantOptions, help="The grant to use", default="device"
    )
    parser.add_argument("--silent", dest="silent", action="store_true")
    parser.add_argument("--refresh", dest="refresh", action="store_true")
    parser.add_argument(
        "--services",
        type=str,
        help="The services you want to connect to (seperated by ,)",
    )
    parser.set_defaults(silent=False)
    parser.set_defaults(refresh=False)
    parser.set_defaults(claim=False)

    args = parser.parse_args()
    print(args.silent)

    main(
        script=args.script,
        path=args.path,
        name=args.name,
        refresh=args.refresh,
        scan_network=args.scan_network,
        template=args.template,
        services=args.services.split(",") if args.services else [],
        silent=args.silent,
        token=args.token,
        tag=args.t,
        endpoint_url=args.endpoint_url,
        grant=args.grant,
        retries=args.retries,
        interval=args.interval,
    )


if __name__ == "__main__":
    entrypoint()



import os
from rich import get_console
from rich.prompt import Confirm, Prompt
from .initialize import load_manifest, ensure_gitignore
from .connect import load_credentials
from arkitekt.cli.dev.autostage import watch_directory_and_stage
import asyncio


def dev(path, silent=False, token=None, endpoint=None):
    """Run the app

    Args:
        app_directory (str): _description_
        silent (bool, optional): _description_. Defaults to False.
    """
    if path == ".":
        app_directory = os.getcwd()
    else:
        app_directory = os.path.join(os.getcwd(), path)

    console = get_console()
    config_path = os.path.join(app_directory, ".arkitekt")
    manifest_path = os.path.join(config_path, "manifest.yaml")
    credentials_path = os.path.join(config_path, "credentials.yaml")

    initialize_fakts = True

    if not os.path.isfile(manifest_path):
        console.print("No manifest found. Please initialize the app first.")
        return

    if not os.path.isfile(credentials_path):
        console.print("No credentials found. Please connect the app first.")
        return

    if not silent:
        console.print(f"Loading manifest")

    if not token or not endpoint:
        credentials = load_credentials(app_directory)
        endpoint, token = next(iter(credentials.items()))

    asyncio.run(watch_directory_and_stage(path, token=token, endpoint=endpoint))



from arkitekt.apps import Arkitekt
from fakts.grants.remote.device_code import DeviceCodeGrant
from fakts.fakts import Fakts
from fakts.grants import CacheGrant
from herre.grants import CacheGrant as HerreCacheGrant
from herre.grants.oauth2.refresh import RefreshGrant
from herre.grants.fakts import FaktsGrant
from fakts.discovery import StaticDiscovery
from herre import Herre
import logging
from rich.logging import RichHandler

logging.basicConfig(level="INFO", handlers=[RichHandler()])
logger = logging.getLogger(__name__)


def easy(
    identifier: str, version: str = "latest", url: str = "http://localhost:8000/f/"
) -> Arkitekt:
    """
    A simple way to create an Arkitekt app with a device code grant
    it will cache the configuration in a file called `identifier`_token.json
    and the cache in a file called `identifier`_cache.json
    """

    return Arkitekt(
        fakts=Fakts(
            grant=CacheGrant(
                cache_file=f"{identifier}-{version}_cache.json",
                hash=f"{identifier}-{version}-{url}",
                grant=DeviceCodeGrant(
                    identifier=identifier,
                    version=version,
                    discovery=StaticDiscovery(base_url=url),
                ),
            )
        ),
        herre=Herre(
            grant=HerreCacheGrant(
                cache_file=f"{identifier}-{version}_herre_cache.json",
                hash=f"{identifier}-{version}-{url}",
                grant=RefreshGrant(grant=FaktsGrant()),
            ),
        ),
    )


def qt(version: str, identifier: str, parent) -> Arkitekt:
    """
    A simple way to create an Arkitekt app within a Qt application stilll
    utilizing a device code grant to authenticate the user on an application
    level
    """

    from koil.composition.qt import QtPedanticKoil

    return Arkitekt(
        koil=QtPedanticKoil(parent=parent),
        fakts=Fakts(
            grant=CacheGrant(
                cache_file=f"{identifier}-{version}_cache.json",
                grant=DeviceCodeGrant(
                    identifier=identifier,
                    version=version,
                ),
            )
        ),
        herre=Herre(
            grant=HerreCacheGrant(
                cache_file=f"{identifier}-{version}_herre_cache.json",
                hash=f"{identifier}-{version}",
                grant=FaktsGrant(),
            ),
        ),
    )


def publiqt(version: str, identifier: str, parent) -> Arkitekt:
    """
    A simple way to create an Arkitekt within a Qt application stilll
    utilizing a device code grant to authenticate the user on an application
    level
    """

    from koil.composition.qt import QtPedanticKoil

    return Arkitekt(
        koil=QtPedanticKoil(parent=parent),
        fakts=Fakts(
            grant=CacheGrant(
                cache_file=f"{identifier}-{version}_cache.json",
                grant=DeviceCodeGrant(
                    identifier=identifier,
                    version=version,
                ),
            )
        ),
        herre=Herre(
            grant=FaktsGrant(),
        ),
    )


def flussi(
    identifier: str, version: str = "latest", url: str = "http://localhost:8000/f/"
) -> "ConnectedFluss":
    """
    A simple way to create an Arkitekt app with a device code grant
    it will cache the configuration in a file called `identifier`_token.json
    and the cache in a file called `identifier`_cache.json
    """

    from arkitekt.apps.fluss import ConnectedFluss

    return ConnectedFluss(
        fakts=Fakts(
            grant=CacheGrant(
                cache_file=f"{identifier}-{version}_cache.json",
                hash=f"{identifier}-{version}-{url}",
                grant=DeviceCodeGrant(
                    identifier=identifier,
                    version=version,
                    discovery=StaticDiscovery(base_url=url),
                ),
            )
        ),
        herre=Herre(
            grant=HerreCacheGrant(
                cache_file=f"{identifier}-{version}_herre_cache.json",
                hash=f"{identifier}-{version}-{url}",
                grant=RefreshGrant(grant=FaktsGrant()),
            ),
        ),
    )



import os
from rich import get_console
from rich.prompt import Confirm, Prompt
from .initialize import load_manifest, ensure_gitignore
import yaml
from typing import Dict
from fakts import Fakts
from fakts.grants.remote import DeviceCodeGrant
from fakts.discovery.base import FaktsEndpoint
from fakts.discovery.static import StaticDiscovery
import asyncio


available_scopes = [
    "arkitekt:rekuest:imitate",
    "arkitekt:herre",
    "arkitekt:fakts",
    "arkitekt:mikro:read",
    "arkitekt:mikro:imitate",
    "arkitekt:unlok",
]


def load_credentials(app_directory: str) -> Dict[str, str]:
    """Load the manifest from the app directory

    Args:
        app_directory (str): _description_

    Returns:
        [type]: _description_
    """
    config_path = os.path.join(app_directory, ".arkitekt")
    c_path = os.path.join(config_path, "credentials.yaml")

    with open(c_path, "r") as f:
        return yaml.safe_load(f)


def connect(
    app_directory: str,
    endpoint: str = None,
    refresh=None,
    silent=False,
):
    """Initialize a new Arkitekt app



    Args:
        console (Console): _description_
        app_directory (str): _description_
        refresh (_type_, optional): _description_. Defaults to None.
        silent (bool, optional): _description_. Defaults to False.
    """

    # Create the project directory

    # Create the arkitekt.json file
    console = get_console()
    config_path = os.path.join(app_directory, ".arkitekt")
    manifest_path = os.path.join(config_path, "manifest.yaml")
    credentials_path = os.path.join(config_path, "credentials.yaml")

    initialize_fakts = True

    if not os.path.isfile(manifest_path):
        console.print("No manifest found. Please initialize the app first.")
        return

    console.print(f"Initializing a new Configuration")

    url = endpoint or (
        Prompt.ask(
            "Give us the path to your fakts instance",
            default=os.getenv("FAKTS_ENDPOINT_URL", "http://localhost:8000/f/"),
        )
        if not silent
        else os.getenv("FAKTS_ENDPOINT_URL", "http://localhost:8000/f/")
    )

    console.print("Connecting to fakts instance at " + url)

    manifest = load_manifest(app_directory)

    grant = DeviceCodeGrant(
        identifier=manifest.identifier,
        version=manifest.version,
        discovery=StaticDiscovery(base_url=url),
    )

    token = asyncio.run(grant.ademand(endpoint=FaktsEndpoint(base_url=url)))

    with open(credentials_path, "w") as f:
        yaml.safe_dump({url: token}, f)

    ensure_gitignore(app_directory)



import os
from rich import get_console
from rich.prompt import Confirm, Prompt
from .initialize import load_manifest, ensure_gitignore
from .connect import load_credentials
import yaml
from fakts import Fakts
from fakts.grants.remote import DeviceCodeGrant
from fakts.discovery.base import FaktsEndpoint
from fakts.discovery.static import StaticDiscovery
from arkitekt.cli.prod.run import import_directory_and_start
import asyncio
import subprocess
from typing import List
from pydantic import BaseModel
import pydantic


class BuildManifest(BaseModel):
    builder: str = "port"
    version: str
    identifier: str
    scopes: list
    image: str


class PortYaml(BaseModel):
    builds: List[BuildManifest]


def build(path, silent=False, tag=None):
    """Run the app

    Args:
        app_directory (str): _description_
        silent (bool, optional): _description_. Defaults to False.
    """
    if path == ".":
        app_directory = os.getcwd()
    else:
        app_directory = os.path.join(os.getcwd(), path)

    assert tag, "Please provide a tag"

    print(app_directory)
    dockerfile = os.path.join(app_directory, "Dockerfile")
    if not os.path.isfile(dockerfile):
        console = get_console()
        console.print("No Dockerfile found. Please create a Dockerfile first.")
        return

    manifest = load_manifest(app_directory)
    print(tag)

    x = subprocess.run(["docker", "build", "-t", tag, "."], cwd=app_directory)

    portyaml = os.path.join(app_directory, ".arkitekt/port.yaml")
    if os.path.isfile(portyaml):
        with open(portyaml, "w") as f:
            try:
                port = PortYaml(**yaml.safe_load(f))
            except Exception as e:
                console = get_console()
                console.print("Invalid port.yaml file. Overwriting")
                port = PortYaml(builds=[])
                return
    else:
        port = PortYaml(builds=[])

    port.builds.append(BuildManifest(**manifest.dict(), image=tag, builder="port"))

    with open(portyaml, "w") as f:
        yaml.safe_dump(port.dict(), f)
